# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Defines a lightweight wrapper class to wrap a single GRIB message.

"""

from collections import namedtuple
import re

import gribapi
import numpy as np
import numpy.ma as ma

from iris._lazy_data import as_lazy_data
from iris.exceptions import TranslationError

_SUPPORTED_GRID_DEFINITIONS = (0, 1, 5, 10, 12, 20, 30, 40, 90, 140)


class _OpenFileRef:
    """
    A reference to an open file that ensures that the file is closed
    when the object is garbage collected.
    """
    def __init__(self, open_file):
        self.open_file = open_file

    def __del__(self):
        if not self.open_file.closed:
            self.open_file.close()


class GribMessage:
    """
    An in-memory representation of a GribMessage, providing
    access to the :meth:`~GribMessage.data` payload and the metadata
    elements by section via the :meth:`~GribMessage.sections` property.

    """

    @staticmethod
    def messages_from_filename(filename):
        """
        Return a generator of :class:`GribMessage` instances; one for
        each message in the supplied GRIB file.

        Args:

        * filename (string):
            Name of the file to generate fields from.

        """
        grib_fh = open(filename, 'rb')
        # create an _OpenFileRef to manage the closure of the file handle
        file_ref = _OpenFileRef(grib_fh)

        while True:
            offset = grib_fh.tell()
            grib_id = gribapi.grib_new_from_file(grib_fh)
            if grib_id is None:
                break
            raw_message = _RawGribMessage(grib_id)
            recreate_raw = _MessageLocation(filename, offset)
            yield GribMessage(raw_message, recreate_raw, file_ref=file_ref)

    def __init__(self, raw_message, recreate_raw, file_ref=None):
        """
        It is recommended to obtain GribMessage instance from the static method
        :meth:`~GribMessage.messages_from_filename`, rather than creating
        them directly.

        """
        # A RawGribMessage giving gribapi access to the original grib message.
        self._raw_message = raw_message
        # A _MessageLocation which dask uses to read the message data array,
        # by which time this message may be dead and the original grib file
        # closed.
        self._recreate_raw = recreate_raw
        # An _OpenFileRef to keep the grib file open while this GribMessage is
        # alive, so that we can always use self._raw_message to fetch keys.
        self._file_ref = file_ref

    @property
    def sections(self):
        """
        Return the key-value pairs of the message keys, grouped by containing
        section.

        Sections in a message are indexed by GRIB section-number,
        and values in a section are indexed by key strings.

        .. For example::

            print(grib_message.sections[4]['parameterNumber'])
            grib_message.sections[1]['minute'] = 0

        """
        return self._raw_message.sections

    @property
    def bmdi(self):
        # Not sure of any cases where GRIB provides a fill value.
        # Default for fill value is None.
        return None

    def core_data(self):
        return self.data

    @property
    def data(self):
        """
        The data array from the GRIB message as a dask Array.

        The shape of the array will match the logical shape of the
        message's grid. For example, a simple global grid would be
        available as a 2-dimensional array with shape (Nj, Ni).

        """
        sections = self.sections
        grid_section = sections[3]
        if grid_section['sourceOfGridDefinition'] != 0:
            raise TranslationError(
                'Unsupported source of grid definition: {}'.format(
                    grid_section['sourceOfGridDefinition']))

        reduced = (grid_section['numberOfOctectsForNumberOfPoints'] != 0 or
                   grid_section['interpretationOfNumberOfPoints'] != 0)
        template = grid_section['gridDefinitionTemplateNumber']
        if reduced and template not in (40,):
            raise TranslationError('Grid definition Section 3 contains '
                                   'unsupported quasi-regular grid.')

        if template in _SUPPORTED_GRID_DEFINITIONS:
            # We can ignore the first two bits (i-neg, j-pos) because
            # that is already captured in the coordinate values.
            if grid_section['scanningMode'] & 0x3f:
                msg = 'Unsupported scanning mode: {}'.format(
                    grid_section['scanningMode'])
                raise TranslationError(msg)
            if template in (20, 30, 90):
                shape = (grid_section['Ny'], grid_section['Nx'])
            elif template == 140:
                shape = (grid_section['numberOfPointsAlongYAxis'],
                         grid_section['numberOfPointsAlongXAxis'])
            elif template == 40 and reduced:
                shape = (grid_section['numberOfDataPoints'],)
            else:
                shape = (grid_section['Nj'], grid_section['Ni'])
            proxy = _DataProxy(shape, np.dtype('f8'), self._recreate_raw)
            data = as_lazy_data(proxy)
        else:
            fmt = 'Grid definition template {} is not supported'
            raise TranslationError(fmt.format(template))
        return data

    def __getstate__(self):
        """
        Alter state of object prior to pickle, ensure open file is closed.

        """
        if not self._file_ref.open_file.closed:
            self._file_ref.open_file.close()
        return self


class _MessageLocation(namedtuple('_MessageLocation', 'filename offset')):
    """A reference to a specific GRIB message within a file."""

    __slots__ = ()

    def __call__(self):
        return _RawGribMessage.from_file_offset(self.filename, self.offset)


class _DataProxy:
    """A reference to the data payload of a single GRIB message."""

    __slots__ = ('shape', 'dtype', 'recreate_raw')

    def __init__(self, shape, dtype, recreate_raw):
        self.shape = shape
        self.dtype = dtype
        self.recreate_raw = recreate_raw

    @property
    def ndim(self):
        return len(self.shape)

    def _bitmap(self, bitmap_section):
        """
        Get the bitmap for the data from the message. The GRIB spec defines
        that the bitmap is composed of values 0 or 1, where:

            * 0: no data value at corresponding data point (data point masked).
            * 1: data value at corresponding data point (data point unmasked).

        The bitmap can take the following values:

            * 0: Bitmap applies to the data and is specified in this section
                 of this message.
            * 1-253: Bitmap applies to the data, is specified by originating
                     centre and is not specified in section 6 of this message.
            * 254: Bitmap applies to the data, is specified in an earlier
                   section 6 of this message and is not specified in this
                   section 6 of this message.
            * 255: Bitmap does not apply to the data.

        Only values 0 and 255 are supported.

        Returns the bitmap as a 1D array of length equal to the
        number of data points in the message.

        """
        # Reference GRIB2 Code Table 6.0.
        bitMapIndicator = bitmap_section['bitMapIndicator']

        if bitMapIndicator == 0:
            bitmap = bitmap_section['bitmap']
        elif bitMapIndicator == 255:
            bitmap = None
        else:
            msg = 'Bitmap Section 6 contains unsupported ' \
                  'bitmap indicator [{}]'.format(bitMapIndicator)
            raise TranslationError(msg)
        return bitmap

    def __getitem__(self, keys):
        # NB. Currently assumes that the validity of this interpretation
        # is checked before this proxy is created.
        message = self.recreate_raw()
        sections = message.sections
        bitmap_section = sections[6]
        bitmap = self._bitmap(bitmap_section)
        data = sections[7]['codedValues']

        if bitmap is not None:
            # Note that bitmap and data are both 1D arrays at this point.
            if np.count_nonzero(bitmap) == data.shape[0]:
                # Only the non-masked values are included in codedValues.
                _data = np.empty(shape=bitmap.shape)
                _data[bitmap.astype(bool)] = data
                # `ma.masked_array` masks where input = 1, the opposite of
                # the behaviour specified by the GRIB spec.
                data = ma.masked_array(_data, mask=np.logical_not(bitmap),
                                       fill_value=np.nan)
            else:
                msg = 'Shapes of data and bitmap do not match.'
                raise TranslationError(msg)

        data = data.reshape(self.shape)
        return data.__getitem__(keys)

    def __repr__(self):
        msg = '<{self.__class__.__name__} shape={self.shape} ' \
            'dtype={self.dtype!r} recreate_raw={self.recreate_raw!r} '
        return msg.format(self=self)

    def __getstate__(self):
        return {attr: getattr(self, attr) for attr in self.__slots__}

    def __setstate__(self, state):
        for key, value in state.items():
            setattr(self, key, value)


class _RawGribMessage:
    """
    Lightweight GRIB message wrapper, containing **only** the coded keys
    of the input GRIB message.

    """
    _NEW_SECTION_KEY_MATCHER = re.compile(r'section([0-9]{1})Length')

    @staticmethod
    def from_file_offset(filename, offset):
        with open(filename, 'rb') as f:
            f.seek(offset)
            message_id = gribapi.grib_new_from_file(f)
            if message_id is None:
                fmt = 'Invalid GRIB message: {} @ {}'
                raise RuntimeError(fmt.format(filename, offset))
        return _RawGribMessage(message_id)

    def __init__(self, message_id):
        """
        A _RawGribMessage object contains the **coded** keys from a
        GRIB message that is identified by the input message id.

        Args:

        * message_id:
            An integer generated by gribapi referencing a GRIB message within
            an open GRIB file.

        """
        self._message_id = message_id
        self._sections = None

    def __del__(self):
        """
        Release the gribapi reference to the message at end of object's life.

        """
        gribapi.grib_release(self._message_id)

    @property
    def sections(self):
        """
        Return the key-value pairs of the message keys, grouped by containing
        section.

        Key-value pairs are collected into a dictionary of
        :class:`Section` objects. One such object is made for
        each section in the message, such that the section number is the
        object's key in the containing dictionary. Each object contains
        key-value pairs for all of the message keys in the given section.

        """
        if self._sections is None:
            self._sections = self._get_message_sections()
        return self._sections

    def _get_message_keys(self):
        """Creates a generator of all the keys in the message."""

        keys_itr = gribapi.grib_keys_iterator_new(self._message_id)
        gribapi.grib_skip_computed(keys_itr)
        while gribapi.grib_keys_iterator_next(keys_itr):
            yield gribapi.grib_keys_iterator_get_name(keys_itr)
        gribapi.grib_keys_iterator_delete(keys_itr)

    def _get_message_sections(self):
        """
        Group keys by section.

        Returns a dictionary mapping section number to :class:`Section`
        instance.

        .. seealso::
            The sections property (:meth:`~sections`).

        """
        sections = {}
        # The first keys in a message are for the whole message and are
        # contained in section 0.
        section = new_section = 0
        section_keys = []

        for key_name in self._get_message_keys():
            # The `section<1-7>Length` keys mark the start of each new
            # section, except for section 8 which is marked by the key '7777'.
            key_match = re.match(self._NEW_SECTION_KEY_MATCHER, key_name)
            if key_match is not None:
                new_section = int(key_match.group(1))
            elif key_name == '7777':
                new_section = 8
            if section != new_section:
                sections[section] = Section(self._message_id, section,
                                            section_keys)
                section_keys = []
                section = new_section
            section_keys.append(key_name)
        sections[section] = Section(self._message_id, section, section_keys)
        return sections


class Section:
    """
    A Section of a GRIB message, supporting dictionary like access to
    attributes using gribapi key strings.

    Values for keys may be changed using assignment but this does not
    write to the file.

    """
    # Keys are read from the file as required and values are cached.
    # Within GribMessage instances all keys will have been fetched

    def __init__(self, message_id, number, keys):
        self._message_id = message_id
        self._number = number
        self._keys = keys
        self._cache = {}

    def __repr__(self):
        items = []
        for key in self._keys:
            value = self._cache.get(key, '?')
            items.append('{}={}'.format(key, value))
        return '<{} {}: {}>'.format(type(self).__name__, self._number,
                                    ', '.join(items))

    def __getitem__(self, key):
        if key not in self._cache:
            if key == 'numberOfSection':
                value = self._number
            elif key not in self._keys:
                raise KeyError('{!r} not defined in section {}'.format(
                    key, self._number))
            else:
                value = self._get_key_value(key)
            self._cache[key] = value
        return self._cache[key]

    def __setitem__(self, key, value):
        # Allow the overwriting of any entry already in the _cache.
        if key in self._cache:
            self._cache[key] = value
        else:
            raise KeyError('{!r} cannot be redefined in '
                           'section {}'.format(key, self._number))

    def _get_key_value(self, key):
        """
        Get the value associated with the given key in the GRIB message.

        Args:

        * key:
            The GRIB key to retrieve the value of.

        Returns the value associated with the requested key in the GRIB
        message.

        """
        vector_keys = ('codedValues', 'pv', 'satelliteSeries',
                       'satelliteNumber', 'instrumentType',
                       'scaleFactorOfCentralWaveNumber',
                       'scaledValueOfCentralWaveNumber',
                       'longitudes', 'latitudes')
        if key in vector_keys:
            res = gribapi.grib_get_array(self._message_id, key)
        elif key == 'bitmap':
            # The bitmap is stored as contiguous boolean bits, one bit for each
            # data point. GRIBAPI returns these as strings, so it must be
            # type-cast to return an array of ints (0, 1).
            res = gribapi.grib_get_array(self._message_id, key, int)
        elif key in ('typeOfFirstFixedSurface', 'typeOfSecondFixedSurface'):
            # By default these values are returned as unhelpful strings but
            # we can use int representation to compare against instead.
            res = self._get_value_or_missing(key, use_int=True)
        else:
            res = self._get_value_or_missing(key)
        return res

    def get_computed_key(self, key):
        """
        Get the computed value associated with the given key in the GRIB
        message.

        Args:

        * key:
            The GRIB key to retrieve the value of.

        Returns the value associated with the requested key in the GRIB
        message.

        """
        vector_keys = ('longitudes', 'latitudes', 'distinctLatitudes')
        if key in vector_keys:
            res = gribapi.grib_get_array(self._message_id, key)
        else:
            res = self._get_value_or_missing(key)
        return res

    def keys(self):
        """Return coded keys available in this Section."""
        return self._keys

    def _get_value_or_missing(self, key, use_int=False):
        """
        Return value of header element, or None if value is encoded as missing.
        Implementation of Regulations 92.1.4 and 92.1.5 via ECCodes.

        """
        if gribapi.grib_is_missing(self._message_id, key):
            result = None
        else:
            if use_int:
                result = gribapi.grib_get(self._message_id, key, int)
            else:
                result = gribapi.grib_get(self._message_id, key)
        return result



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris_grib._load_convert.product_definition_template_11`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from copy import deepcopy
from unittest import mock
import warnings

from iris.coords import DimCoord, CellMethod

from iris_grib._load_convert import product_definition_template_11


class Test(tests.IrisGribTest):
    def setUp(self):
        def func(s, m, f):
            return m['cell_methods'].append(self.cell_method)

        module = 'iris_grib._load_convert'
        self.patch('warnings.warn')
        this_module = '{}.product_definition_template_11'.format(module)
        self.cell_method = mock.sentinel.cell_method
        self.patch(this_module, side_effect=func)
        self.patch_statistical_fp_coord = self.patch(
            module + '.statistical_forecast_period_coord',
            return_value=mock.sentinel.dummy_fp_coord)
        self.patch_time_coord = self.patch(
            module + '.validity_time_coord',
            return_value=mock.sentinel.dummy_time_coord)
        self.patch_vertical_coords = self.patch(module + '.vertical_coords')
        self.metadata = {'factories': [], 'references': [],
                         'standard_name': None,
                         'long_name': None, 'units': None, 'attributes': {},
                         'cell_methods': [], 'dim_coords_and_dims': [],
                         'aux_coords_and_dims': []}

    def _check(self, request_warning):
        grib_lconv_opt = 'iris_grib._load_convert.options'
        with mock.patch(grib_lconv_opt, warn_on_unsupported=request_warning):
            metadata = deepcopy(self.metadata)
            perturbationNumber = 666
            section = {'productDefinitionTemplateNumber': 11,
                       'perturbationNumber': perturbationNumber,
                       'hoursAfterDataCutoff': 1,
                       'minutesAfterDataCutoff': 1,
                       'numberOfTimeRange': 1,
                       'typeOfStatisticalProcessing': 1,
                       'typeOfTimeIncrement': 2,
                       'timeIncrement': 0,
                       'yearOfEndOfOverallTimeInterval': 2000,
                       'monthOfEndOfOverallTimeInterval': 1,
                       'dayOfEndOfOverallTimeInterval': 1,
                       'hourOfEndOfOverallTimeInterval': 1,
                       'minuteOfEndOfOverallTimeInterval': 0,
                       'secondOfEndOfOverallTimeInterval': 1}
            forecast_reference_time = mock.Mock()
            # The called being tested.
            product_definition_template_11(section, metadata,
                                           forecast_reference_time)
            expected = {'cell_methods': [], 'aux_coords_and_dims': []}
            expected['cell_methods'].append(CellMethod(method='sum',
                                                       coords=('time',)))
            realization = DimCoord(perturbationNumber,
                                   standard_name='realization',
                                   units='no_unit')
            expected['aux_coords_and_dims'].append((realization, None))
            self.maxDiff = None
            self.assertEqual(metadata['aux_coords_and_dims'][-1],
                             expected['aux_coords_and_dims'][0])
            self.assertEqual(metadata['cell_methods'][-1],
                             expected['cell_methods'][0])

            if request_warning:
                warn_msgs = [mcall[1][0] for mcall in warnings.warn.mock_calls]
                expected_msgs = ['type of ensemble', 'number of forecasts']
                for emsg in expected_msgs:
                    matches = [wmsg for wmsg in warn_msgs if emsg in wmsg]
                    self.assertEqual(len(matches), 1)
                    warn_msgs.remove(matches[0])
            else:
                self.assertEqual(len(warnings.warn.mock_calls), 0)

    def test_pdt_no_warn(self):
        self._check(False)

    def test_pdt_warn(self):
        self._check(True)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests for function :func:`iris_grib._load_convert.translate_phenomenon`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from copy import deepcopy

from cf_units import Unit
from iris.coords import DimCoord

from iris_grib._load_convert import Probability, translate_phenomenon
from iris_grib.grib_phenom_translation import _GribToCfDataClass, GRIBCode


class Test_probability(tests.IrisGribTest):
    def setUp(self):
        # Patch inner call to return a given phenomenon type.
        target_module = 'iris_grib._load_convert'
        self.phenom_lookup_patch = self.patch(
            target_module + '.itranslation.grib2_phenom_to_cf_info',
            return_value=_GribToCfDataClass('air_temperature', '', 'K', None))
        # Construct dummy call arguments
        self.probability = Probability('<prob_type>', 22.0)
        self.metadata = {'aux_coords_and_dims': [], 'attributes': {}}

    def test_basic(self):
        translate_phenomenon(self.metadata, 7, 8, 9, None,
                             None, None, probability=self.probability)
        # Check metadata.
        thresh_coord = DimCoord([22.0],
                                standard_name='air_temperature',
                                long_name='', units='K')
        self.assertEqual(self.metadata, {
            'standard_name': None,
            'long_name': 'probability_of_air_temperature_<prob_type>',
            'units': Unit(1),
            'aux_coords_and_dims': [(thresh_coord, None)],
            'attributes': {'GRIB_PARAM': GRIBCode(2, 7, 8, 9)}})

    def test_no_phenomenon(self):
        self.phenom_lookup_patch.return_value = None
        expected_metadata = self.metadata.copy()
        translate_phenomenon(self.metadata,
                             discipline=7,
                             parameterCategory=77,
                             parameterNumber=777,
                             typeOfFirstFixedSurface=None,
                             scaledValueOfFirstFixedSurface=None,
                             typeOfSecondFixedSurface=None,
                             probability=self.probability)
        expected_metadata['attributes']['GRIB_PARAM'] = \
            GRIBCode(2, 7, 77, 777)
        self.assertEqual(self.metadata, expected_metadata)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris_grib._load_convert.projection centre.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from iris_grib._load_convert import projection_centre, ProjectionCentre


class Test(tests.IrisGribTest):
    def test_unset(self):
        expected = ProjectionCentre(False, False)
        self.assertEqual(projection_centre(0x0), expected)

    def test_bipolar_and_symmetric(self):
        expected = ProjectionCentre(False, True)
        self.assertEqual(projection_centre(0x40), expected)

    def test_south_pole_on_projection_plane(self):
        expected = ProjectionCentre(True, False)
        self.assertEqual(projection_centre(0x80), expected)

    def test_both(self):
        expected = ProjectionCentre(True, True)
        self.assertEqual(projection_centre(0xc0), expected)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :meth:`iris_grib._save_rules.grid_definition_template_12`.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

import numpy as np

import iris.coords
from iris.coord_systems import GeogCS, TransverseMercator
from iris.exceptions import TranslationError

from iris_grib._save_rules import grid_definition_template_12
from iris_grib.tests.unit.save_rules import GdtTestMixin


class FakeGribError(Exception):
    pass


class Test(tests.IrisGribTest, GdtTestMixin):
    def setUp(self):
        self.default_ellipsoid = GeogCS(semi_major_axis=6377563.396,
                                        semi_minor_axis=6356256.909)
        self.test_cube = self._make_test_cube()

        GdtTestMixin.setUp(self)

    def _make_test_cube(self, cs=None, x_points=None, y_points=None):
        # Create a cube with given properties, or minimal defaults.
        if cs is None:
            cs = self._default_coord_system()
        if x_points is None:
            x_points = self._default_x_points()
        if y_points is None:
            y_points = self._default_y_points()

        x_coord = iris.coords.DimCoord(x_points, 'projection_x_coordinate',
                                       units='m', coord_system=cs)
        y_coord = iris.coords.DimCoord(y_points, 'projection_y_coordinate',
                                       units='m', coord_system=cs)
        test_cube = iris.cube.Cube(np.zeros((len(y_points), len(x_points))))
        test_cube.add_dim_coord(y_coord, 0)
        test_cube.add_dim_coord(x_coord, 1)
        return test_cube

    def _default_coord_system(self):
        # This defines an OSGB coord system.
        cs = TransverseMercator(latitude_of_projection_origin=49.0,
                                longitude_of_central_meridian=-2.0,
                                false_easting=400000.0,
                                false_northing=-100000.0,
                                scale_factor_at_central_meridian=0.9996012717,
                                ellipsoid=self.default_ellipsoid)
        return cs

    def test__template_number(self):
        grid_definition_template_12(self.test_cube, self.mock_grib)
        self._check_key('gridDefinitionTemplateNumber', 12)

    def test__shape_of_earth(self):
        grid_definition_template_12(self.test_cube, self.mock_grib)
        self._check_key('shapeOfTheEarth', 7)
        self._check_key('scaleFactorOfEarthMajorAxis', 0)
        self._check_key('scaledValueOfEarthMajorAxis', 6377563.396)
        self._check_key('scaleFactorOfEarthMinorAxis', 0)
        self._check_key('scaledValueOfEarthMinorAxis', 6356256.909)

    def test__grid_shape(self):
        test_cube = self._make_test_cube(x_points=np.arange(13),
                                         y_points=np.arange(6))
        grid_definition_template_12(test_cube, self.mock_grib)
        self._check_key('Ni', 13)
        self._check_key('Nj', 6)

    def test__grid_points_exact(self):
        test_cube = self._make_test_cube(x_points=[1, 3, 5, 7],
                                         y_points=[4, 9])
        grid_definition_template_12(test_cube, self.mock_grib)
        self._check_key("X1", 100)
        self._check_key("X2", 700)
        self._check_key("Y1", 400)
        self._check_key("Y2", 900)
        self._check_key("Di", 200)
        self._check_key("Dj", 500)

    def test__grid_points_approx(self):
        test_cube = self._make_test_cube(x_points=[1.001, 3.003, 5.005, 7.007],
                                         y_points=[4, 9])
        grid_definition_template_12(test_cube, self.mock_grib)
        self._check_key("X1", 100)
        self._check_key("X2", 701)
        self._check_key("Y1", 400)
        self._check_key("Y2", 900)
        self._check_key("Di", 200)
        self._check_key("Dj", 500)

    def test__negative_grid_points_gribapi_broken(self):
        self.mock_gribapi.GribInternalError = FakeGribError

        # Force the test to run the signed int --> unsigned int workaround.
        def set(grib, key, value):
            if key in ["X1", "X2", "Y1", "Y2"] and value < 0:
                raise self.mock_gribapi.GribInternalError()
            grib.keys[key] = value
        self.mock_gribapi.grib_set = set

        test_cube = self._make_test_cube(x_points=[-1, 1, 3, 5, 7],
                                         y_points=[-4, 9])
        grid_definition_template_12(test_cube, self.mock_grib)
        self._check_key("X1", 0x80000064)
        self._check_key("X2", 700)
        self._check_key("Y1", 0x80000190)
        self._check_key("Y2", 900)

    def test__negative_grid_points_gribapi_fixed(self):
        test_cube = self._make_test_cube(x_points=[-1, 1, 3, 5, 7],
                                         y_points=[-4, 9])
        grid_definition_template_12(test_cube, self.mock_grib)
        self._check_key("X1", -100)
        self._check_key("X2", 700)
        self._check_key("Y1", -400)
        self._check_key("Y2", 900)

    def test__template_specifics(self):
        grid_definition_template_12(self.test_cube, self.mock_grib)
        self._check_key("latitudeOfReferencePoint", 49000000.0)
        self._check_key("longitudeOfReferencePoint", -2000000.0)
        self._check_key("XR", 40000000.0)
        self._check_key("YR", -10000000.0)

    def test__scale_factor_gribapi_broken(self):
        # GRIBAPI expects a signed int for scaleFactorAtReferencePoint
        # but it should accept a float, so work around this.
        # See https://software.ecmwf.int/issues/browse/SUP-1100

        def get_native_type(grib, key):
            assert key == "scaleFactorAtReferencePoint"
            return int
        self.mock_gribapi.grib_get_native_type = get_native_type
        grid_definition_template_12(self.test_cube, self.mock_grib)
        self._check_key("scaleFactorAtReferencePoint", 1065346526)

    def test__scale_factor_gribapi_fixed(self):

        def get_native_type(grib, key):
            assert key == "scaleFactorAtReferencePoint"
            return float
        self.mock_gribapi.grib_get_native_type = get_native_type
        grid_definition_template_12(self.test_cube, self.mock_grib)
        self._check_key("scaleFactorAtReferencePoint", 0.9996012717)

    def test__scanmode(self):
        grid_definition_template_12(self.test_cube, self.mock_grib)
        self._check_key('iScansPositively', 1)
        self._check_key('jScansPositively', 1)

    def test__scanmode_reverse(self):
        test_cube = self._make_test_cube(x_points=np.arange(7, 0, -1))
        grid_definition_template_12(test_cube, self.mock_grib)
        self._check_key('iScansPositively', 0)
        self._check_key('jScansPositively', 1)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests for function :func:`iris_grib._load_convert.statistical_forecast_period`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import datetime
from unittest import mock

from iris_grib._load_convert import statistical_forecast_period_coord


class Test(tests.IrisGribTest):
    def setUp(self):
        module = 'iris_grib._load_convert'
        self.module = module
        self.patch_hindcast = self.patch(module + '._hindcast_fix')
        self.forecast_seconds = 0.0
        self.forecast_units = mock.Mock()
        self.forecast_units.convert = lambda x, y: self.forecast_seconds
        self.patch(module + '.time_range_unit',
                   return_value=self.forecast_units)
        self.frt_coord = mock.Mock()
        self.frt_coord.points = [1]
        self.frt_coord.units.num2date = mock.Mock(
            return_value=datetime.datetime(2010, 2, 3))
        self.section = {}
        self.section['yearOfEndOfOverallTimeInterval'] = 2010
        self.section['monthOfEndOfOverallTimeInterval'] = 2
        self.section['dayOfEndOfOverallTimeInterval'] = 3
        self.section['hourOfEndOfOverallTimeInterval'] = 8
        self.section['minuteOfEndOfOverallTimeInterval'] = 0
        self.section['secondOfEndOfOverallTimeInterval'] = 0
        self.section['forecastTime'] = mock.Mock()
        self.section['indicatorOfUnitOfTimeRange'] = mock.Mock()

    def test_basic(self):
        coord = statistical_forecast_period_coord(self.section,
                                                  self.frt_coord)
        self.assertEqual(coord.standard_name, 'forecast_period')
        self.assertEqual(coord.units, 'hours')
        self.assertArrayAlmostEqual(coord.points, [4.0])
        self.assertArrayAlmostEqual(coord.bounds, [[0.0, 8.0]])

    def test_with_hindcast(self):
        coord = statistical_forecast_period_coord(self.section,
                                                  self.frt_coord)
        self.assertEqual(self.patch_hindcast.call_count, 1)

    def test_no_hindcast(self):
        self.patch(self.module + '.options.support_hindcast_values', False)
        coord = statistical_forecast_period_coord(self.section,
                                                  self.frt_coord)
        self.assertEqual(self.patch_hindcast.call_count, 0)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for `iris_grib._save_rules.fixup_float32_as_int32`.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from iris_grib._save_rules import fixup_float32_as_int32


class Test(tests.IrisGribTest):
    def test_positive_zero(self):
        result = fixup_float32_as_int32(0.0)
        self.assertEqual(result, 0)

    def test_negative_zero(self):
        result = fixup_float32_as_int32(-0.0)
        self.assertEqual(result, 0)

    def test_high_bit_clear_1(self):
        # Start with the float32 value for the bit pattern 0x00000001.
        result = fixup_float32_as_int32(1.401298464324817e-45)
        self.assertEqual(result, 1)

    def test_high_bit_clear_2(self):
        # Start with the float32 value for the bit pattern 0x00000002.
        result = fixup_float32_as_int32(2.802596928649634e-45)
        self.assertEqual(result, 2)

    def test_high_bit_set_1(self):
        # Start with the float32 value for the bit pattern 0x80000001.
        result = fixup_float32_as_int32(-1.401298464324817e-45)
        self.assertEqual(result, -1)

    def test_high_bit_set_2(self):
        # Start with the float32 value for the bit pattern 0x80000002.
        result = fixup_float32_as_int32(-2.802596928649634e-45)
        self.assertEqual(result, -2)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for `iris_grib.grib_save_rules.identification`."""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from unittest import mock

import gribapi

import iris
import iris.tests.stock as stock

from iris_grib._save_rules import identification
from iris_grib.tests.unit import TestGribSimple


GRIB_API = 'iris_grib._save_rules.gribapi'


class Test(TestGribSimple):
    @tests.skip_data
    def test_no_realization(self):
        cube = stock.simple_pp()
        grib = mock.Mock()
        mock_gribapi = mock.Mock(spec=gribapi)
        with mock.patch(GRIB_API, mock_gribapi):
            identification(cube, grib)

        mock_gribapi.assert_has_calls(
            [mock.call.grib_set_long(grib, "typeOfProcessedData", 2)])

    @tests.skip_data
    def test_realization_0(self):
        cube = stock.simple_pp()
        realisation = iris.coords.AuxCoord((0,), standard_name='realization',
                                           units='1')
        cube.add_aux_coord(realisation)

        grib = mock.Mock()
        mock_gribapi = mock.Mock(spec=gribapi)
        with mock.patch(GRIB_API, mock_gribapi):
            identification(cube, grib)

        mock_gribapi.assert_has_calls(
            [mock.call.grib_set_long(grib, "typeOfProcessedData", 3)])

    @tests.skip_data
    def test_realization_n(self):
        cube = stock.simple_pp()
        realisation = iris.coords.AuxCoord((2,), standard_name='realization',
                                           units='1')
        cube.add_aux_coord(realisation)

        grib = mock.Mock()
        mock_gribapi = mock.Mock(spec=gribapi)
        with mock.patch(GRIB_API, mock_gribapi):
            identification(cube, grib)

        mock_gribapi.assert_has_calls(
            [mock.call.grib_set_long(grib, "typeOfProcessedData", 4)])


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function
:func:`iris_grib._load_convert.grid_definition_template_5`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from copy import deepcopy
from unittest import mock

from iris_grib._load_convert import grid_definition_template_5


class Test(tests.IrisGribTest):
    def setUp(self):
        def func(s, m, y, x, c):
            return m['dim_coords_and_dims'].append(item)

        module = 'iris_grib._load_convert'

        self.major = mock.sentinel.major
        self.minor = mock.sentinel.minor
        self.radius = mock.sentinel.radius

        mfunc = '{}.ellipsoid_geometry'.format(module)
        return_value = (self.major, self.minor, self.radius)
        self.patch(mfunc, return_value=return_value)

        mfunc = '{}.ellipsoid'.format(module)
        self.ellipsoid = mock.sentinel.ellipsoid
        self.patch(mfunc, return_value=self.ellipsoid)

        mfunc = '{}.grid_definition_template_4_and_5'.format(module)
        self.coord = mock.sentinel.coord
        self.dim = mock.sentinel.dim
        item = (self.coord, self.dim)
        self.patch(mfunc, side_effect=func)

        mclass = 'iris.coord_systems.RotatedGeogCS'
        self.cs = mock.sentinel.cs
        self.patch(mclass, return_value=self.cs)

        self.metadata = {'factories': [], 'references': [],
                         'standard_name': None,
                         'long_name': None, 'units': None, 'attributes': {},
                         'cell_methods': [], 'dim_coords_and_dims': [],
                         'aux_coords_and_dims': []}

    def test(self):
        metadata = deepcopy(self.metadata)
        angleOfRotation = mock.sentinel.angleOfRotation
        shapeOfTheEarth = mock.sentinel.shapeOfTheEarth
        section = {'latitudeOfSouthernPole': 45000000,
                   'longitudeOfSouthernPole': 90000000,
                   'angleOfRotation': angleOfRotation,
                   'shapeOfTheEarth': shapeOfTheEarth}
        # The called being tested.
        grid_definition_template_5(section, metadata)
        from iris_grib._load_convert import \
            ellipsoid_geometry, \
            ellipsoid, \
            grid_definition_template_4_and_5 as gdt_4_5
        self.assertEqual(ellipsoid_geometry.call_count, 1)
        ellipsoid.assert_called_once_with(shapeOfTheEarth, self.major,
                                          self.minor, self.radius)
        from iris.coord_systems import RotatedGeogCS
        RotatedGeogCS.assert_called_once_with(-45.0, 270.0, angleOfRotation,
                                              self.ellipsoid)
        gdt_4_5.assert_called_once_with(section, metadata, 'grid_latitude',
                                        'grid_longitude', self.cs)
        expected = deepcopy(self.metadata)
        expected['dim_coords_and_dims'].append((self.coord, self.dim))
        self.assertEqual(metadata, expected)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`iris_grib._load_convert.grib2_convert`."""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import copy
from unittest import mock

import iris_grib
from iris_grib._load_convert import grib2_convert
from iris_grib.tests.unit import _make_test_message


class Test(tests.IrisGribTest):
    def setUp(self):
        this = 'iris_grib._load_convert'
        self.patch('{}.reference_time_coord'.format(this), return_value=None)
        self.patch('{}.grid_definition_section'.format(this))
        self.patch('{}.product_definition_section'.format(this))
        self.patch('{}.data_representation_section'.format(this))
        self.patch('{}.bitmap_section'.format(this))

    def test(self):
        sections = [{'discipline': mock.sentinel.discipline},       # section 0
                    {'centre': 'ecmf',                              # section 1
                     'tablesVersion': mock.sentinel.tablesVersion},
                    None,                                           # section 2
                    mock.sentinel.grid_definition_section,          # section 3
                    mock.sentinel.product_definition_section,       # section 4
                    mock.sentinel.data_representation_section,      # section 5
                    mock.sentinel.bitmap_section]                   # section 6
        field = _make_test_message(sections)
        metadata = {'factories': [], 'references': [],
                    'standard_name': None,
                    'long_name': None, 'units': None, 'attributes': {},
                    'cell_methods': [], 'dim_coords_and_dims': [],
                    'aux_coords_and_dims': []}
        expected = copy.deepcopy(metadata)
        centre = 'European Centre for Medium Range Weather Forecasts'
        expected['attributes'] = {'centre': centre}
        # The call being tested.
        grib2_convert(field, metadata)
        self.assertEqual(metadata, expected)
        this = iris_grib._load_convert
        this.reference_time_coord.assert_called_with(sections[1])
        this.grid_definition_section.assert_called_with(sections[3],
                                                        expected)
        args = (sections[4], expected, sections[0]['discipline'],
                sections[1]['tablesVersion'], None)
        this.product_definition_section.assert_called_with(*args)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Integration tests for NAME to GRIB2 interoperability."""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

import numpy as np
import warnings

import iris


def name_cb(cube, field, filename):
    # NAME files give the time point at the end of the range but Iris'
    # GRIB loader creates it in the middle (the GRIB file itself doesn't
    # encode a time point). Here we make them consistent so we can
    # easily compare them.
    t_coord = cube.coord("time")
    t_coord.points = t_coord.bounds[0][1]
    fp_coord = cube.coord("forecast_period")
    fp_coord.points = fp_coord.bounds[0][1]
    # NAME contains extra vertical meta-data.
    z_coord = cube.coords("height")
    if z_coord:
        z_coord[0].standard_name = "height"
        z_coord[0].long_name = "height above ground level"


class TestNameToGRIB(tests.IrisGribTest):
    def check_common(self, name_cube, grib_cube):
        self.assertTrue(np.allclose(name_cube.data, grib_cube.data))
        self.assertTrue(
            np.allclose(
                name_cube.coord("latitude").points,
                grib_cube.coord("latitude").points,
            )
        )
        self.assertTrue(
            np.allclose(
                name_cube.coord("longitude").points,
                grib_cube.coord("longitude").points - 360,
            )
        )

        for c in ["height", "time"]:
            if name_cube.coords(c):
                self.assertEqual(name_cube.coord(c), grib_cube.coord(c))

    @tests.skip_data
    def test_name2_field(self):
        filepath = tests.get_data_path(("NAME", "NAMEII_field.txt"))
        name_cubes = iris.load(filepath)

        # There is a known load/save problem with numerous
        # gribapi/eccodes versions and
        # zero only data, where min == max.
        # This may be a problem with data scaling.
        for i, name_cube in enumerate(name_cubes):
            data = name_cube.data
            if np.min(data) == np.max(data):
                msg = (
                    'NAMEII cube #{}, "{}" has empty data : '
                    "SKIPPING test for this cube, as save/load will "
                    "not currently work."
                )
                warnings.warn(msg.format(i, name_cube.name()))
                continue

            with self.temp_filename(".grib2") as temp_filename:
                iris.save(name_cube, temp_filename)
                grib_cube = iris.load_cube(temp_filename, callback=name_cb)
                self.check_common(name_cube, grib_cube)
                self.assertCML(
                    grib_cube,
                    self.get_result_path(
                        (
                            "integration",
                            "name_grib",
                            "NAMEII",
                            "{}_{}.cml".format(i, name_cube.name()),
                        )
                    ),
                )

    @tests.skip_data
    def test_name3_field(self):
        filepath = tests.get_data_path(("NAME", "NAMEIII_field.txt"))
        name_cubes = iris.load(filepath)
        for i, name_cube in enumerate(name_cubes):
            with self.temp_filename(".grib2") as temp_filename:
                iris.save(name_cube, temp_filename)
                grib_cube = iris.load_cube(temp_filename, callback=name_cb)

                self.check_common(name_cube, grib_cube)
                self.assertCML(
                    grib_cube,
                    self.get_result_path(
                        (
                            "integration",
                            "name_grib",
                            "NAMEIII",
                            "{}_{}.cml".format(i, name_cube.name()),
                        )
                    ),
                )


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
#
# DO NOT EDIT: AUTO-GENERATED
# Created on 29 November 2019 10:11 from 
# http://www.metarelate.net/metOcean
# at commit 36da82fe65a1a37e6cf1fa49899107fc6dc908ed

# https://github.com/metarelate/metOcean/commit/36da82fe65a1a37e6cf1fa49899107fc6dc908ed

"""
Provides GRIB/CF phenomenon translations.

"""

from collections import namedtuple


CFName = namedtuple('CFName', 'standard_name long_name units')

DimensionCoordinate = namedtuple('DimensionCoordinate',
                                 'standard_name units points')

G1LocalParam = namedtuple('G1LocalParam', 'edition t2version centre iParam')
G2Param = namedtuple('G2Param', 'edition discipline category number')


GRIB1_LOCAL_TO_CF_CONSTRAINED = {
    G1LocalParam(1, 128, 98, 165): (CFName('x_wind', None, 'm s-1'), DimensionCoordinate('height', 'm', (10,))),
    G1LocalParam(1, 128, 98, 166): (CFName('y_wind', None, 'm s-1'), DimensionCoordinate('height', 'm', (10,))),
    G1LocalParam(1, 128, 98, 167): (CFName('air_temperature', None, 'K'), DimensionCoordinate('height', 'm', (2,))),
    G1LocalParam(1, 128, 98, 168): (CFName('dew_point_temperature', None, 'K'), DimensionCoordinate('height', 'm', (2,))),
    }

GRIB1_LOCAL_TO_CF = {
    G1LocalParam(1, 128, 98, 31): CFName('sea_ice_area_fraction', None, '1'),
    G1LocalParam(1, 128, 98, 34): CFName('sea_surface_temperature', None, 'K'),
    G1LocalParam(1, 128, 98, 59): CFName('atmosphere_specific_convective_available_potential_energy', None, 'J kg-1'),
    G1LocalParam(1, 128, 98, 129): CFName('geopotential', None, 'm2 s-2'),
    G1LocalParam(1, 128, 98, 130): CFName('air_temperature', None, 'K'),
    G1LocalParam(1, 128, 98, 131): CFName('x_wind', None, 'm s-1'),
    G1LocalParam(1, 128, 98, 132): CFName('y_wind', None, 'm s-1'),
    G1LocalParam(1, 128, 98, 135): CFName('lagrangian_tendency_of_air_pressure', None, 'Pa s-1'),
    G1LocalParam(1, 128, 98, 141): CFName('thickness_of_snowfall_amount', None, 'm'),
    G1LocalParam(1, 128, 98, 151): CFName('air_pressure_at_sea_level', None, 'Pa'),
    G1LocalParam(1, 128, 98, 157): CFName('relative_humidity', None, '%'),
    G1LocalParam(1, 128, 98, 164): CFName('cloud_area_fraction', None, '1'),
    G1LocalParam(1, 128, 98, 173): CFName('surface_roughness_length', None, 'm'),
    G1LocalParam(1, 128, 98, 174): CFName(None, 'grib_physical_atmosphere_albedo', '1'),
    G1LocalParam(1, 128, 98, 186): CFName('low_type_cloud_area_fraction', None, '1'),
    G1LocalParam(1, 128, 98, 187): CFName('medium_type_cloud_area_fraction', None, '1'),
    G1LocalParam(1, 128, 98, 188): CFName('high_type_cloud_area_fraction', None, '1'),
    G1LocalParam(1, 128, 98, 235): CFName(None, 'grib_skin_temperature', 'K'),
    }

GRIB2_TO_CF = {
    G2Param(2, 0, 0, 0): CFName('air_temperature', None, 'K'),
    G2Param(2, 0, 0, 2): CFName('air_potential_temperature', None, 'K'),
    G2Param(2, 0, 0, 6): CFName('dew_point_temperature', None, 'K'),
    G2Param(2, 0, 0, 10): CFName('surface_upward_latent_heat_flux', None, 'W m-2'),
    G2Param(2, 0, 0, 11): CFName('surface_upward_sensible_heat_flux', None, 'W m-2'),
    G2Param(2, 0, 0, 17): CFName('surface_temperature', None, 'K'),
    G2Param(2, 0, 1, 0): CFName('specific_humidity', None, 'kg kg-1'),
    G2Param(2, 0, 1, 1): CFName('relative_humidity', None, '%'),
    G2Param(2, 0, 1, 2): CFName('humidity_mixing_ratio', None, 'kg kg-1'),
    G2Param(2, 0, 1, 3): CFName(None, 'precipitable_water', 'kg m-2'),
    G2Param(2, 0, 1, 7): CFName('precipitation_flux', None, 'kg m-2 s-1'),
    G2Param(2, 0, 1, 11): CFName('thickness_of_snowfall_amount', None, 'm'),
    G2Param(2, 0, 1, 13): CFName('liquid_water_content_of_surface_snow', None, 'kg m-2'),
    G2Param(2, 0, 1, 22): CFName(None, 'cloud_mixing_ratio', 'kg kg-1'),
    G2Param(2, 0, 1, 49): CFName('precipitation_amount', None, 'kg m-2'),
    G2Param(2, 0, 1, 51): CFName('atmosphere_mass_content_of_water', None, 'kg m-2'),
    G2Param(2, 0, 1, 53): CFName('snowfall_flux', None, 'kg m-2 s-1'),
    G2Param(2, 0, 1, 60): CFName('snowfall_amount', None, 'kg m-2'),
    G2Param(2, 0, 1, 64): CFName('atmosphere_mass_content_of_water_vapor', None, 'kg m-2'),
    G2Param(2, 0, 1, 83): CFName('mass_fraction_of_cloud_liquid_water_in_air', None, 'kg kg-1'),
    G2Param(2, 0, 1, 84): CFName('mass_fraction_of_cloud_ice_in_air', None, 'kg kg-1'),
    G2Param(2, 0, 2, 0): CFName('wind_from_direction', None, 'degrees'),
    G2Param(2, 0, 2, 1): CFName('wind_speed', None, 'm s-1'),
    G2Param(2, 0, 2, 2): CFName('x_wind', None, 'm s-1'),
    G2Param(2, 0, 2, 3): CFName('y_wind', None, 'm s-1'),
    G2Param(2, 0, 2, 8): CFName('lagrangian_tendency_of_air_pressure', None, 'Pa s-1'),
    G2Param(2, 0, 2, 10): CFName('atmosphere_absolute_vorticity', None, 's-1'),
    G2Param(2, 0, 2, 14): CFName(None, 'ertel_potential_velocity', 'K m2 kg-1 s-1'),
    G2Param(2, 0, 2, 22): CFName('wind_speed_of_gust', None, 'm s-1'),
    G2Param(2, 0, 3, 0): CFName('air_pressure', None, 'Pa'),
    G2Param(2, 0, 3, 1): CFName('air_pressure_at_sea_level', None, 'Pa'),
    G2Param(2, 0, 3, 3): CFName(None, 'icao_standard_atmosphere_reference_height', 'm'),
    G2Param(2, 0, 3, 4): CFName('geopotential', None, 'm2 s-2'),
    G2Param(2, 0, 3, 5): CFName('geopotential_height', None, 'm'),
    G2Param(2, 0, 3, 6): CFName('altitude', None, 'm'),
    G2Param(2, 0, 3, 9): CFName('geopotential_height_anomaly', None, 'm'),
    G2Param(2, 0, 4, 7): CFName('surface_downwelling_shortwave_flux_in_air', None, 'W m-2'),
    G2Param(2, 0, 4, 9): CFName('surface_net_downward_shortwave_flux', None, 'W m-2'),
    G2Param(2, 0, 5, 3): CFName('surface_downwelling_longwave_flux_in_air', None, 'W m-2'),
    G2Param(2, 0, 5, 5): CFName('surface_net_downward_longwave_flux', None, 'W m-2'),
    G2Param(2, 0, 6, 1): CFName(None, 'cloud_area_fraction_assuming_maximum_random_overlap', '1'),
    G2Param(2, 0, 6, 3): CFName('low_type_cloud_area_fraction', None, '%'),
    G2Param(2, 0, 6, 4): CFName('medium_type_cloud_area_fraction', None, '%'),
    G2Param(2, 0, 6, 5): CFName('high_type_cloud_area_fraction', None, '%'),
    G2Param(2, 0, 6, 6): CFName('atmosphere_mass_content_of_cloud_liquid_water', None, 'kg m-2'),
    G2Param(2, 0, 6, 7): CFName('cloud_area_fraction_in_atmosphere_layer', None, '%'),
    G2Param(2, 0, 6, 25): CFName(None, 'WAFC_CB_horizontal_extent', '1'),
    G2Param(2, 0, 6, 26): CFName(None, 'WAFC_ICAO_height_at_cloud_base', 'm'),
    G2Param(2, 0, 6, 27): CFName(None, 'WAFC_ICAO_height_at_cloud_top', 'm'),
    G2Param(2, 0, 7, 6): CFName('atmosphere_specific_convective_available_potential_energy', None, 'J kg-1'),
    G2Param(2, 0, 7, 7): CFName(None, 'convective_inhibition', 'J kg-1'),
    G2Param(2, 0, 7, 8): CFName(None, 'storm_relative_helicity', 'J kg-1'),
    G2Param(2, 0, 14, 0): CFName('atmosphere_mole_content_of_ozone', None, 'Dobson'),
    G2Param(2, 0, 19, 1): CFName(None, 'grib_physical_atmosphere_albedo', '%'),
    G2Param(2, 0, 19, 20): CFName(None, 'WAFC_icing_potential', '1'),
    G2Param(2, 0, 19, 21): CFName(None, 'WAFC_in-cloud_turb_potential', '1'),
    G2Param(2, 0, 19, 22): CFName(None, 'WAFC_CAT_potential', '1'),
    G2Param(2, 2, 0, 0): CFName('land_binary_mask', None, '1'),
    G2Param(2, 2, 0, 0): CFName('land_area_fraction', None, '1'),
    G2Param(2, 2, 0, 1): CFName('surface_roughness_length', None, 'm'),
    G2Param(2, 2, 0, 2): CFName('soil_temperature', None, 'K'),
    G2Param(2, 2, 0, 7): CFName('surface_altitude', None, 'm'),
    G2Param(2, 2, 0, 22): CFName('moisture_content_of_soil_layer', None, 'kg m-2'),
    G2Param(2, 2, 0, 34): CFName('surface_runoff_flux', None, 'kg m-2 s-1'),
    G2Param(2, 10, 1, 2): CFName('sea_water_x_velocity', None, 'm s-1'),
    G2Param(2, 10, 1, 3): CFName('sea_water_y_velocity', None, 'm s-1'),
    G2Param(2, 10, 2, 0): CFName('sea_ice_area_fraction', None, '1'),
    G2Param(2, 10, 3, 0): CFName('sea_surface_temperature', None, 'K'),
    }

CF_CONSTRAINED_TO_GRIB1_LOCAL = {
    (CFName('air_temperature', None, 'K'), DimensionCoordinate('height', 'm', (2,))): G1LocalParam(1, 128, 98, 167),
    (CFName('dew_point_temperature', None, 'K'), DimensionCoordinate('height', 'm', (2,))): G1LocalParam(1, 128, 98, 168),
    (CFName('x_wind', None, 'm s-1'), DimensionCoordinate('height', 'm', (10,))): G1LocalParam(1, 128, 98, 165),
    (CFName('y_wind', None, 'm s-1'), DimensionCoordinate('height', 'm', (10,))): G1LocalParam(1, 128, 98, 166),
    }

CF_TO_GRIB1_LOCAL = {
    CFName(None, 'grib_physical_atmosphere_albedo', '1'): G1LocalParam(1, 128, 98, 174),
    CFName(None, 'grib_skin_temperature', 'K'): G1LocalParam(1, 128, 98, 235),
    CFName('air_pressure_at_sea_level', None, 'Pa'): G1LocalParam(1, 128, 98, 151),
    CFName('air_temperature', None, 'K'): G1LocalParam(1, 128, 98, 130),
    CFName('atmosphere_specific_convective_available_potential_energy', None, 'J kg-1'): G1LocalParam(1, 128, 98, 59),
    CFName('cloud_area_fraction', None, '1'): G1LocalParam(1, 128, 98, 164),
    CFName('geopotential', None, 'm2 s-2'): G1LocalParam(1, 128, 98, 129),
    CFName('high_type_cloud_area_fraction', None, '1'): G1LocalParam(1, 128, 98, 188),
    CFName('lagrangian_tendency_of_air_pressure', None, 'Pa s-1'): G1LocalParam(1, 128, 98, 135),
    CFName('low_type_cloud_area_fraction', None, '1'): G1LocalParam(1, 128, 98, 186),
    CFName('medium_type_cloud_area_fraction', None, '1'): G1LocalParam(1, 128, 98, 187),
    CFName('relative_humidity', None, '%'): G1LocalParam(1, 128, 98, 157),
    CFName('sea_ice_area_fraction', None, '1'): G1LocalParam(1, 128, 98, 31),
    CFName('sea_surface_temperature', None, 'K'): G1LocalParam(1, 128, 98, 34),
    CFName('surface_roughness_length', None, 'm'): G1LocalParam(1, 128, 98, 173),
    CFName('thickness_of_snowfall_amount', None, 'm'): G1LocalParam(1, 128, 98, 141),
    CFName('x_wind', None, 'm s-1'): G1LocalParam(1, 128, 98, 131),
    CFName('y_wind', None, 'm s-1'): G1LocalParam(1, 128, 98, 132),
    }

CF_TO_GRIB2 = {
    CFName(None, 'WAFC_CAT_potential', '1'): G2Param(2, 0, 19, 22),
    CFName(None, 'WAFC_CB_horizontal_extent', '1'): G2Param(2, 0, 6, 25),
    CFName(None, 'WAFC_ICAO_height_at_cloud_base', 'm'): G2Param(2, 0, 6, 26),
    CFName(None, 'WAFC_ICAO_height_at_cloud_top', 'm'): G2Param(2, 0, 6, 27),
    CFName(None, 'WAFC_icing_potential', '1'): G2Param(2, 0, 19, 20),
    CFName(None, 'WAFC_in-cloud_turb_potential', '1'): G2Param(2, 0, 19, 21),
    CFName(None, 'cloud_area_fraction_assuming_maximum_random_overlap', '1'): G2Param(2, 0, 6, 1),
    CFName(None, 'cloud_mixing_ratio', 'kg kg-1'): G2Param(2, 0, 1, 22),
    CFName(None, 'convective_inhibition', 'J kg-1'): G2Param(2, 0, 7, 7),
    CFName(None, 'ertel_potential_velocity', 'K m2 kg-1 s-1'): G2Param(2, 0, 2, 14),
    CFName(None, 'grib_physical_atmosphere_albedo', '%'): G2Param(2, 0, 19, 1),
    CFName(None, 'icao_standard_atmosphere_reference_height', 'm'): G2Param(2, 0, 3, 3),
    CFName(None, 'precipitable_water', 'kg m-2'): G2Param(2, 0, 1, 3),
    CFName(None, 'storm_relative_helicity', 'J kg-1'): G2Param(2, 0, 7, 8),
    CFName('air_potential_temperature', None, 'K'): G2Param(2, 0, 0, 2),
    CFName('air_pressure', None, 'Pa'): G2Param(2, 0, 3, 0),
    CFName('air_pressure_at_sea_level', None, 'Pa'): G2Param(2, 0, 3, 0),
    CFName('air_pressure_at_sea_level', None, 'Pa'): G2Param(2, 0, 3, 1),
    CFName('air_temperature', None, 'K'): G2Param(2, 0, 0, 0),
    CFName('altitude', None, 'm'): G2Param(2, 0, 3, 6),
    CFName('atmosphere_absolute_vorticity', None, 's-1'): G2Param(2, 0, 2, 10),
    CFName('atmosphere_mass_content_of_cloud_liquid_water', None, 'kg m-2'): G2Param(2, 0, 6, 6),
    CFName('atmosphere_mass_content_of_water', None, 'kg m-2'): G2Param(2, 0, 1, 51),
    CFName('atmosphere_mass_content_of_water_vapor', None, 'kg m-2'): G2Param(2, 0, 1, 64),
    CFName('atmosphere_mole_content_of_ozone', None, 'Dobson'): G2Param(2, 0, 14, 0),
    CFName('atmosphere_specific_convective_available_potential_energy', None, 'J kg-1'): G2Param(2, 0, 7, 6),
    CFName('cloud_area_fraction_in_atmosphere_layer', None, '%'): G2Param(2, 0, 6, 7),
    CFName('dew_point_temperature', None, 'K'): G2Param(2, 0, 0, 6),
    CFName('geopotential', None, 'm2 s-2'): G2Param(2, 0, 3, 4),
    CFName('geopotential_height', None, 'm'): G2Param(2, 0, 3, 5),
    CFName('geopotential_height_anomaly', None, 'm'): G2Param(2, 0, 3, 9),
    CFName('high_type_cloud_area_fraction', None, '%'): G2Param(2, 0, 6, 5),
    CFName('humidity_mixing_ratio', None, 'kg kg-1'): G2Param(2, 0, 1, 2),
    CFName('lagrangian_tendency_of_air_pressure', None, 'Pa s-1'): G2Param(2, 0, 2, 8),
    CFName('land_area_fraction', None, '1'): G2Param(2, 2, 0, 0),
    CFName('land_binary_mask', None, '1'): G2Param(2, 2, 0, 0),
    CFName('liquid_water_content_of_surface_snow', None, 'kg m-2'): G2Param(2, 0, 1, 13),
    CFName('low_type_cloud_area_fraction', None, '%'): G2Param(2, 0, 6, 3),
    CFName('mass_fraction_of_cloud_ice_in_air', None, 'kg kg-1'): G2Param(2, 0, 1, 84),
    CFName('mass_fraction_of_cloud_liquid_water_in_air', None, 'kg kg-1'): G2Param(2, 0, 1, 83),
    CFName('medium_type_cloud_area_fraction', None, '%'): G2Param(2, 0, 6, 4),
    CFName('moisture_content_of_soil_layer', None, 'kg m-2'): G2Param(2, 2, 0, 22),
    CFName('precipitation_amount', None, 'kg m-2'): G2Param(2, 0, 1, 49),
    CFName('precipitation_flux', None, 'kg m-2 s-1'): G2Param(2, 0, 1, 7),
    CFName('relative_humidity', None, '%'): G2Param(2, 0, 1, 1),
    CFName('sea_ice_area_fraction', None, '1'): G2Param(2, 10, 2, 0),
    CFName('sea_surface_temperature', None, 'K'): G2Param(2, 10, 3, 0),
    CFName('sea_water_x_velocity', None, 'm s-1'): G2Param(2, 10, 1, 2),
    CFName('sea_water_y_velocity', None, 'm s-1'): G2Param(2, 10, 1, 3),
    CFName('snowfall_amount', None, 'kg m-2'): G2Param(2, 0, 1, 60),
    CFName('snowfall_flux', None, 'kg m-2 s-1'): G2Param(2, 0, 1, 53),
    CFName('soil_temperature', None, 'K'): G2Param(2, 2, 0, 2),
    CFName('specific_humidity', None, 'kg kg-1'): G2Param(2, 0, 1, 0),
    CFName('surface_air_pressure', None, 'Pa'): G2Param(2, 0, 3, 0),
    CFName('surface_altitude', None, 'm'): G2Param(2, 2, 0, 7),
    CFName('surface_downwelling_longwave_flux_in_air', None, 'W m-2'): G2Param(2, 0, 5, 3),
    CFName('surface_downwelling_shortwave_flux_in_air', None, 'W m-2'): G2Param(2, 0, 4, 7),
    CFName('surface_net_downward_longwave_flux', None, 'W m-2'): G2Param(2, 0, 5, 5),
    CFName('surface_net_downward_shortwave_flux', None, 'W m-2'): G2Param(2, 0, 4, 9),
    CFName('surface_roughness_length', None, 'm'): G2Param(2, 2, 0, 1),
    CFName('surface_runoff_flux', None, 'kg m-2 s-1'): G2Param(2, 2, 0, 34),
    CFName('surface_temperature', None, 'K'): G2Param(2, 0, 0, 17),
    CFName('surface_upward_latent_heat_flux', None, 'W m-2'): G2Param(2, 0, 0, 10),
    CFName('surface_upward_sensible_heat_flux', None, 'W m-2'): G2Param(2, 0, 0, 11),
    CFName('thickness_of_snowfall_amount', None, 'm'): G2Param(2, 0, 1, 11),
    CFName('wind_from_direction', None, 'degrees'): G2Param(2, 0, 2, 0),
    CFName('wind_speed', None, 'm s-1'): G2Param(2, 0, 2, 1),
    CFName('wind_speed_of_gust', None, 'm s-1'): G2Param(2, 0, 2, 22),
    CFName('x_wind', None, 'm s-1'): G2Param(2, 0, 2, 2),
    CFName('y_wind', None, 'm s-1'): G2Param(2, 0, 2, 3),
    }



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris_grib._load_convert.product_definition_template_40`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import iris.coords

from iris_grib._load_convert import product_definition_template_40, _MDI
from iris_grib.tests.unit.load_convert import empty_metadata


class Test(tests.IrisGribTest):
    def setUp(self):
        self.section_4 = {'hoursAfterDataCutoff': _MDI,
                          'minutesAfterDataCutoff': _MDI,
                          'constituentType': 1,
                          'indicatorOfUnitOfTimeRange': 0,  # minutes
                          'startStep': 360,
                          'NV': 0,
                          'typeOfFirstFixedSurface': 103,
                          'scaleFactorOfFirstFixedSurface': 0,
                          'scaledValueOfFirstFixedSurface': 9999,
                          'typeOfSecondFixedSurface': 255}

    def test_constituent_type(self):
        metadata = empty_metadata()
        rt_coord = iris.coords.DimCoord(24, 'forecast_reference_time',
                                        units='hours since epoch')
        product_definition_template_40(self.section_4, metadata, rt_coord)
        expected = empty_metadata()
        expected['attributes']['WMO_constituent_type'] = 1
        self.assertEqual(metadata['attributes'], expected['attributes'])


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :func:`iris_grib._save_rules.data_section`.

"""

# import iris_grib.tests first so that some things can be initialised before
# importing anything else
import iris_grib.tests as tests

from unittest import mock

import gribapi
import iris.cube
import numpy as np

from iris_grib._save_rules import data_section


GRIB_API = 'iris_grib._save_rules.gribapi'
GRIB_MESSAGE = mock.sentinel.GRIB_MESSAGE


class TestMDI(tests.IrisGribTest):
    def assertBitmapOff(self, grib_api):
        # Check the use of a mask has been turned off via:
        #   gribapi.grib_set(grib_message, 'bitmapPresent', 0)
        grib_api.grib_set.assert_called_once_with(GRIB_MESSAGE,
                                                  'bitmapPresent', 0)

    def assertBitmapOn(self, grib_api, fill_value):
        # Check the use of a mask has been turned on via:
        #   gribapi.grib_set(grib_message, 'bitmapPresent', 1)
        #   gribapi.grib_set_double(grib_message, 'missingValue', fill_value)
        grib_api.grib_set.assert_called_once_with(GRIB_MESSAGE,
                                                  'bitmapPresent', 1)
        grib_api.grib_set_double.assert_called_once_with(GRIB_MESSAGE,
                                                         'missingValue',
                                                         fill_value)

    def assertBitmapRange(self, grib_api, min_data, max_data):
        # Check the use of a mask has been turned on via:
        #   gribapi.grib_set(grib_message, 'bitmapPresent', 1)
        #   gribapi.grib_set_double(grib_message, 'missingValue', ...)
        # and that a suitable fill value has been chosen.
        grib_api.grib_set.assert_called_once_with(GRIB_MESSAGE,
                                                  'bitmapPresent', 1)
        args, = grib_api.grib_set_double.call_args_list
        (message, key, fill_value), kwargs = args
        self.assertIs(message, GRIB_MESSAGE)
        self.assertEqual(key, 'missingValue')
        self.assertTrue(fill_value < min_data or fill_value > max_data,
                        'Fill value {} is not outside data range '
                        '{} to {}.'.format(fill_value, min_data, max_data))
        return fill_value

    def assertValues(self, grib_api, values):
        # Check the correct data values have been set via:
        #   gribapi.grib_set_double_array(grib_message, 'values', ...)
        args, = grib_api.grib_set_double_array.call_args_list
        (message, key, values), kwargs = args
        self.assertIs(message, GRIB_MESSAGE)
        self.assertEqual(key, 'values')
        self.assertArrayEqual(values, values)
        self.assertEqual(kwargs, {})

    def test_simple(self):
        # Check the simple case of non-masked data with no scaling.
        cube = iris.cube.Cube(np.arange(5))
        grib_message = mock.sentinel.GRIB_MESSAGE
        with mock.patch(GRIB_API) as grib_api:
            data_section(cube, grib_message)
        # Check the use of a mask has been turned off.
        self.assertBitmapOff(grib_api)
        # Check the correct data values have been set.
        self.assertValues(grib_api, np.arange(5))

    def test_masked_with_finite_fill_value(self):
        cube = iris.cube.Cube(np.ma.MaskedArray([1.0, 2.0, 3.0, 1.0, 2.0, 3.0],
                                                mask=[0, 0, 0, 1, 1, 1],
                                                fill_value=2000))
        grib_message = mock.sentinel.GRIB_MESSAGE
        with mock.patch(GRIB_API) as grib_api:
            data_section(cube, grib_message)
        # Check the use of a mask has been turned on.
        FILL = 2000
        self.assertBitmapOn(grib_api, FILL)
        # Check the correct data values have been set.
        self.assertValues(grib_api, [1, 2, 3, FILL, FILL, FILL])

    def test_masked_with_nan_fill_value(self):
        cube = iris.cube.Cube(np.ma.MaskedArray([1.0, 2.0, 3.0, 1.0, 2.0, 3.0],
                                                mask=[0, 0, 0, 1, 1, 1],
                                                fill_value=np.nan))
        grib_message = mock.sentinel.GRIB_MESSAGE
        with mock.patch(GRIB_API) as grib_api:
            data_section(cube, grib_message)
        # Check the use of a mask has been turned on and a suitable fill
        # value has been chosen.
        FILL = self.assertBitmapRange(grib_api, 1, 3)
        # Check the correct data values have been set.
        self.assertValues(grib_api, [1, 2, 3, FILL, FILL, FILL])

    def test_scaled(self):
        # If the Cube's units don't match the units required by GRIB
        # ensure the data values are scaled correctly.
        cube = iris.cube.Cube(np.arange(5),
                              standard_name='geopotential_height', units='km')
        grib_message = mock.sentinel.GRIB_MESSAGE
        with mock.patch(GRIB_API) as grib_api:
            data_section(cube, grib_message)
        # Check the use of a mask has been turned off.
        self.assertBitmapOff(grib_api)
        # Check the correct data values have been set.
        self.assertValues(grib_api, np.arange(5) * 1000)

    def test_scaled_with_finite_fill_value(self):
        # When re-scaling masked data with a finite fill value, ensure
        # the fill value and any filled values are also re-scaled.
        cube = iris.cube.Cube(np.ma.MaskedArray([1.0, 2.0, 3.0, 1.0, 2.0, 3.0],
                                                mask=[0, 0, 0, 1, 1, 1],
                                                fill_value=2000),
                              standard_name='geopotential_height', units='km')
        grib_message = mock.sentinel.GRIB_MESSAGE
        with mock.patch(GRIB_API) as grib_api:
            data_section(cube, grib_message)
        # Check the use of a mask has been turned on.
        FILL = 2000 * 1000
        self.assertBitmapOn(grib_api, FILL)
        # Check the correct data values have been set.
        self.assertValues(grib_api, [1000, 2000, 3000, FILL, FILL, FILL])

    def test_scaled_with_nan_fill_value(self):
        # When re-scaling masked data with a NaN fill value, ensure
        # a fill value is chosen which allows for the scaling, and any
        # filled values match the chosen fill value.
        cube = iris.cube.Cube(np.ma.MaskedArray([-1.0, 2.0, -1.0, 2.0],
                                                mask=[0, 0, 1, 1],
                                                fill_value=np.nan),
                              standard_name='geopotential_height', units='km')
        grib_message = mock.sentinel.GRIB_MESSAGE
        with mock.patch(GRIB_API) as grib_api:
            data_section(cube, grib_message)
        # Check the use of a mask has been turned on and a suitable fill
        # value has been chosen.
        FILL = self.assertBitmapRange(grib_api, -1000, 2000)
        # Check the correct data values have been set.
        self.assertValues(grib_api, [-1000, 2000, FILL, FILL])


class TestNonDoubleData(tests.IrisGribTest):
    # When saving to GRIB, data that is not float64 is cast to float64. This
    # test checks that non-float64 data is saved without raising a segmentation
    # fault.
    def check(self, dtype):
        data = np.random.random(1920 * 2560).astype(dtype)
        cube = iris.cube.Cube(data,
                              standard_name='geopotential_height', units='km')
        grib_message = gribapi.grib_new_from_samples("GRIB2")
        data_section(cube, grib_message)
        gribapi.grib_release(grib_message)

    def test_float32(self):
        self.check(dtype=np.float32)

    def test_int32(self):
        self.check(dtype=np.int32)

    def test_int64(self):
        self.check(dtype=np.int64)


if __name__ == "__main__":
    tests.main()
